package calc;



/** This class is the base for all kind of calculations. It's subclasses 
represent the nodes in the Operational Tree constructed by ExprParser */

public abstract class BaseCalc {

  protected Complex result;
  protected BaseCalc[] children = null;
  protected boolean calculated = false;

  /** overridden by inheriting class, needed for consistency-checks */
  public abstract int getArgsNr();

  /** connects the Object to another BaseCalc and uses it's output as input */
  public void insertChildren(BaseCalc[] children) throws IllegalArgException{
    if(children.length != getArgsNr()) 
      throw new IllegalArgException("Illegal Number of Arguments to "
				    +this.getClass().getName()+"!\nneeds "
				    +getArgsNr()+" - got "+ children.length);
    this.children = children;
  }

  /** returns an array with the child-nodes */
  public BaseCalc[] getChildren(){
    return children;
  }

  /** Wrapper for doCalculate. Keeps track of the state */
  public void calculate(){
    if(calculated)
      return;
    //System.out.println(getClass().getName()+": calculating...");
    doCalculate();
    calculated = true;
  }

  /** implemented by subclasses, does the real calculation by recursively 
      calling it's children */
  protected abstract void doCalculate();
    
  /** return the result calculated by calculate() */
  public Complex result(){
    if(!calculated)
      calculate();	    
    //System.out.println("result is "+result);
    return result;
  }

  /** remove the calculated-flag */
  public void clear(){
    if(children != null)
      for(int i = 0; i < children.length; i++)
	children[i].clear();
    calculated = false;
  }
    
  /** returns a string representing the arithmetic operation */
  public String toString(){
    String retVal = "";
    if(children != null){
      for(int i=0;i<children.length;i++)
	retVal += " "+children[i].toString();
      return getArithmeticSymbol()+"("+retVal+" )";
    }
    return getArithmeticSymbol();
  }
	

  /** must be implemented for debugging purposes: returns "+" for AddCalc,
      etc. */
  public abstract String getArithmeticSymbol();

  /** for testing purposes */
  public static void main(String[] args){

    // make some nodes:
    MultiplyCalc m = new MultiplyCalc();
    AddCalc a = new AddCalc();
    NumeralCalc n1 = new NumeralCalc(new Complex(1,2));
    NumeralCalc n2 = new NumeralCalc(new Complex(3,4));
    NumeralCalc n3 = new NumeralCalc(new Complex(1,0));

    // construct the Operational Tree: n1 * n2 + n3
    try{
      BaseCalc[] args1 = {n1,n2};
      m.insertChildren(args1);
      BaseCalc[] args2 = {m,n3};
      a.insertChildren(args2);
    } catch (IllegalArgException e){ System.out.println(""+e); }
    System.out.println("Result is: "+a.result());
  }
}

