package calc;


/** The Representation of a Complex Number */

public class Complex implements Comparable {


  static class Debug {
    private static void out(String str) {
      //System.out.println(str);
    }
  }

  public double real = 0,imag = 0;

  /** the "constant" for the Number of partial Integrations in the
      Euler-MacLaurin-Formula used in zeta and zeta_fast */
  public static int Q = 6;

  /** the "constant" for the upper border of the integral (= number of steps
      in which the integration is performed) used in zeta  */
  public static int N = 37;

  
  public final static Complex ZERO = new Complex(0,0);
  public final static Complex ONE = new Complex(1,0);


  public Complex (double real, double imag){

    this.real = real;
    this.imag = imag;
  }

  public Complex (Complex z1){

    this.real = z1.real;
    this.imag = z1.imag;
  }

  public double abs(){
    return  Math.sqrt(real * real + imag * imag);
  }
	
	public double phi(){
		return Math.atan(imag/real);
	}

  public double square(){
    return  real * real + imag * imag;
  }

  public void negate(){
    real = -real;
    imag = - imag;
  }

  public void add(Complex z2){

    real += z2.real;
    imag  += z2.imag;
  }

  public void subtract(Complex z2){

    real -= z2.real;
    imag  -= z2.imag;
  }

  public void mult(Complex z2){
    double newreal = real * z2.real - imag * z2.imag;
    imag = real * z2.imag + imag * z2.real;
    real = newreal;
  }

  public void div(Complex z2){

    double denom = (z2.real * z2.real + z2.imag * z2.imag);
    double newreal = (real * z2.real + imag * z2.imag) / denom;
    imag = (imag * z2.real - real * z2.imag) / denom;
    real = newreal;
  }

  public String toString(){
    return "["+real+" , "+imag+"i]";
  }

  /** we compare the absolute values of complex numbers (needed for sorting)
   this leads to the inconsistency of ordering with equality (which is normally
   assumed):
   a.compareTo(b) = 0  =/=>  a.equals(b) !*/
  public int compareTo(Object o){
    if(abs() - ((Complex)o).abs() > 0)
      return 1;
    if(abs() - ((Complex)o).abs() < 0)
      return -1;
    else
      return 0;
    
  }

  /** two complex numbers are equal, if real- and imaginary part are equal */
  public boolean equals(Object o){
    if(o instanceof Complex &&
       ((Complex)o).real == real && ((Complex)o).imag == imag)
      return true;
    return false;
  }

  /** needed by HashTables (e.g. in ComplexPowerSum.multiply()) */
  public int hashCode(){
    Double realPart = new Double(real);
    Double imagPart = new Double(imag);
    return realPart.hashCode() + imagPart.hashCode();
  }

  /** returns false, if any part of c is NaN (not a number) */
  public boolean isValid(){
    if(Double.isInfinite(real) || Double.isInfinite(real)
       || Double.isInfinite(real) || Double.isNaN(imag))
      return false;
    return true;
  }

  /* the static methods are used, when a new Object of Type Complex is
      needed */

  /** the absolute value */
  public static double abs(Complex z1){
    return Math.sqrt(z1.real * z1.real + z1.imag * z1.imag);
			
  }

	/** the angle phi, when the number is expressed in polar coordinates */
	public static double phi(Complex z1){
		return Math.atan(z1.imag/z1.real);
	}
	
  /** z -> -z */
  public static Complex negate(Complex z1){
    return new Complex(-z1.real,-z1.imag);
  }

  /** z1, z2 -> z1 + z2 */
  public static Complex add(Complex z1, Complex z2){

    return new Complex(z1.real + z2.real,
		       z1.imag + z2.imag);
  }


  /** z1, z2 -> z1 - z2 */
  public static Complex subtract(Complex z1, Complex z2){
    return new Complex(z1.real - z2.real,
		       z1.imag - z2.imag);
  }

  /** z1, z2 -> z1 * z2 */
  public static Complex mult(Complex z1, Complex z2){

    return new Complex(z1.real * z2.real - z1.imag * z2.imag,
		       z1.real * z2.imag + z1.imag * z2.real);
  }

  /** z1, z2 -> z1 / z2 */
  public static Complex div(Complex z1, Complex z2){

    double denom = (z2.real * z2.real + z2.imag * z2.imag);

    return new Complex((z1.real * z2.real + z1.imag * z2.imag)/ denom,
		       (z1.imag * z2.real - z1.real * z2.imag)/ denom);
  }

  public static Complex sin(Complex z1){
	
    return new Complex(Math.sin(z1.real)/2 * (Math.exp(z1.imag)
					      + Math.exp(-z1.imag)),
		       Math.cos(z1.real)/2 * (Math.exp(z1.imag)
					      - Math.exp(-z1.imag)));
  }

  public static Complex cos(Complex z1){
	
    return new Complex(Math.cos(z1.real)/2 * (Math.exp(z1.imag)
					      + Math.exp(-z1.imag)),
		       Math.sin(z1.real)/2 * (Math.exp(-z1.imag)
					      - Math.exp(z1.imag)));
  }

  public static Complex exp(Complex z1){
    return new Complex(Math.exp(z1.real)*Math.cos(z1.imag),
		       Math.exp(z1.real)*Math.sin(z1.imag));
  }
	  

  /** the general power function */
  public static Complex pow(Complex z1, Complex z2){
   
    return exp(mult(z2,ln(z1)));
  }


  /** the natural logarithm */
  public static Complex ln(Complex z1){
    return new Complex(Math.log(abs(z1)), Math.atan(z1.imag/z1.real));
  }

  /** for usability, does the same as ln */
  public static Complex log(Complex z1){
    return ln(z1);
  }


  /** the riemann-zeta-function, "new and improved" calculation. Uses the
      calculation algorithm implemented in PARI */

  public static Complex zeta_fastest(Complex s){
    // determine the lower border of the integral
    int M = (int)Math.ceil(0.27 * Math.abs(s.imag) +4.08);

    Complex C_M = new Complex(M,0);

    // step 1:
    //
    //          _ M
    //         \
    // ... +   /      n^(-s)
    //         -- n=1
    Complex step1 = new Complex(0,0);
    for(int n=M; n >=1; n--)
      step1.add(pow(new Complex(n,0), negate(s)));

    // step 2:
    //
    // (s+1)(s+2)          / (s+3)(s+4)            /   2B_q\     B_4 \   B_2
    // ------------m^(-2)*| ------------ m^(-2)...|0 + --- |...+ --- | + ---
    //   2 * 3             \   4 * 5               \    q  /      2  /    1

    Complex m_2 = mult(C_M, C_M);
    Complex step2 = new Complex(0,0);
    for(int i = Q; i >= 1; i--){
      int i2 = i << 1;
      Complex factor = add(new Complex(i2 - 1, 0), s);
      factor.mult(add(new Complex(i2 + 1, 0), s));
      factor.div(new Complex(i2*(i2+1),0));
      factor.div(m_2);
      step2.mult(factor);
      step2.add(new
		Complex(calc.zeta.BernoulliPolynomBuilder.getBernoulliNumber(i2)/i,
			0));
    }
    step2.mult(s);
    step2.div(new Complex(2*M,0));
    step2.add(div(C_M, subtract(s, ONE)));
    step2.subtract(new Complex(0.5,0));
    step2.mult(pow(C_M, negate(s)));
    step2.add(step1);
    return step2;
  }

  /** the riemann-zeta-function, improved calculation */

  public static Complex zeta_fast(Complex s){

    // determine the lower border of the integral
    int M = (int)Math.ceil(0.27 * Math.abs(s.imag) +4.08);

    Complex C_M = new Complex(M,0);

    // step 1:
    //
    //          _ M
    //         \
    // ... +   /      n^(-s)
    //         -- n=1
    Complex step1 = new Complex(0,0);

    for(int n=M; n >=1; n--)
      step1.add(pow(new Complex(n,0), negate(s)));
    //System.out.println("step1 is: "+step1);
    // step 2:
    //
    //
    //
    // ... + m^(-s+1)/(s-1) - m^(-s)/2)
    //
    Complex step2 = subtract(div(pow(C_M, add(negate(s),
					      ONE)),
			subtract(s,ONE)),
			    div(pow(C_M, negate(s)),
				new Complex(2,0)));

    //System.out.println("step2 is: "+step2);
    step2.add(step1);
				
    // step 3:
    //
    //          _ q
    //         \
    // ... +   /       B_r/r! s * (s+1) *...* (s+r-2)* m^(-s-r+1)
    //         -- r=2

    Complex sum = new Complex(0, 0);
    double b_r;

    // B_{2n+1} == 0, so we only need to sum for the even values
    for(int r = 2; r <= Q; r+=2){

      // product s*(s+1)*..*(s+r-2)
      Complex factor = new Complex(s);
      for(int i = 1; i <= r-2; i++)
	factor.mult(add(s, new Complex(i,0)));

      factor.mult(new
		  Complex(calc.zeta.BernoulliPolynomBuilder.getBernoulliNumber(r), 0));
      factor.div(new Complex(calc.zeta.BernoulliPolynomBuilder.fac(r),0));
      factor.mult(pow(C_M, add(negate(s), new Complex(-r+1,0))));
      sum.add(factor);
    }
    //System.out.println("sum is: "+sum);
    
    Complex step3 = add(step2, sum);
    //System.out.println("step3 is: "+step3);

    return step3;

  }
  /** the riemann-zeta-function */

  public static Complex zeta(Complex s){

    //long time = System.currentTimeMillis();

    // step 1 :
    //
    //  1/(s-1) +1/2
    //

    Complex step1 = add(div(ONE,
			    add(s, new Complex(-1,0))),
			new Complex(0.5d,0));

    //System.out.println("step1 is:"+step1);
    // step 2:
    //
    //          _ q
    //         \
    // ... +   /       B_r/r! s * (s+1) *...* (s+r-2) ... -
    //         -- r=2

    Complex sum = new Complex(0, 0);
    double b_r;
    for(int r = 2; r <= Q; r++){

      // product s*(s+1)*..*(s+r-2)
      Complex factor = new Complex(s);
      for(int i = 1; i <= r-2; i++)
	factor.mult(add(s, new Complex(i,0)));

      factor.mult(new
		  Complex(calc.zeta.BernoulliPolynomBuilder.getBernoulliNumber(r), 0));
      factor.div(new Complex(calc.zeta.BernoulliPolynomBuilder.fac(r),0));

      sum.add(factor);
    }

    Complex step2 = add(step1, sum);
    //System.out.println("step2 is:"+step2);

    // step 3:
    //
    //
    // ... +  1/q! * s * (s+1) *...* (s + q - 1) * (step4)
    //
    //
    
    
    Complex step3 = new Complex(s);
    for(int i = 1; i <= Q-1; i++){
      step3 = mult(step3, add(s, new Complex(i,0)));
    }
    step3.div(new Complex(calc.zeta.BernoulliPolynomBuilder.fac(Q),0));
    step3.negate();
    //System.out.println("step3 is: "+step3);

    // step4:
    //
    //        \inf                          _ \inf   k+1
    //        /                             \       /
    //        | B_q(x -  [x]) x^(-s-q) dx = /       | B_q(x - k) x^(-s-q) dx
    //        /                             -- k=1  /
    //        1                                     k

    Complex step4 = new Complex(0,0);

    try {
      for(int k = 1; k <= N; k++){
	
	// build the polynomial
	Polynom B_q = calc.zeta.BernoulliPolynomBuilder.buildPolynom(Q);
	
	// translate by -k
	B_q.translate((double) -k);
	//System.out.println("BernoulliPolynom is "+B_q.toString()+"!");
	
	// transform to ComplexPowerSum (a Sum of complex coefficients with
	// complex powers)
	ComplexPowerSum B = new ComplexPowerSum(B_q);
	
	
	// create x^(Re(-s)-q)
	Complex[] coefficient = {new Complex(1,0)};
	Complex[] power = {new Complex(- (s.real) - (double) Q, -(s.imag))};
	
	ComplexPowerSum x = new ComplexPowerSum(coefficient, power);
	
	// perform B_q(x - k) * x^(Re(-s)-q)
	//System.out.println("multiplying with: "+x);
	x.multiply(B);
	//System.out.println("\n\nB_q(x -k) * x^(-s-q) is: "+x.toString());
	
	// integrate the result with integration constant zero
	x.integrate(ZERO);
	//System.out.println("\n\nIntegral is: "+x.toString());
	
	//System.out.println("f("+k+") is :"+x.eval(new Complex(k+1,0)));
	// insert integration range
	Complex summand = Complex.subtract(x.eval(new Complex(k+1,0)),
				       x.eval(new Complex(k,0)));
	//System.out.println("summand for "+k+" is: "+summand);
	step4.add(summand);
      }
      //System.out.println("step4 is: "+step4);
    } catch(InvalidException e){ e.printStackTrace();}

    //step5 : multiply with step3 and add step2
    Complex step5 = add(mult(step3,step4),step2);

    //time = System.currentTimeMillis() - time;
    //System.out.println ("Done in "+time/1000 +" seconds.");

    return step5;
  }
	
  /** only for testing purposes */
  public static void main(String args[]){
    if(args.length <4)
      System.out.println("needs 4 args: real1 imag1 real2 imag2");
	
    Complex c1 = new Complex(Double.valueOf(args[0]).doubleValue(),
			     Double.valueOf(args[1]).doubleValue());
    Complex c2 = new Complex(Double.valueOf(args[2]).doubleValue(),
			     Double.valueOf(args[3]).doubleValue());

    System.out.println("c1 = "+c1+"\nc2 = "+c2);

    System.out.println("c1 / c2 = "+ Complex.div(c1,c2));
    System.out.println("c1 * c2 = "+ Complex.mult(c1,c2));

    System.out.println("c1 + c2 = "+Complex.add(c1,c2));
    System.out.println("c1 - c2 = "+Complex.add(c1,Complex.negate(c2)));
    System.out.println("c1 ^ c2 = "+Complex.pow(c1,c2));
    System.out.println("ln c1 = "+Complex.ln(c1));
    System.out.println("exp (ln c1) = "+Complex.exp(ln(c1)));
    System.out.println("exp c1) = "+Complex.exp(c1));
    System.out.println("zeta (c1) = "+Complex.zeta_fastest(c1).abs());
    //System.out.println("zeta (c1) = "+Complex.zeta_fast(c1).abs());
  }

}
