package calc;

import java.util.*;

/* This class represents a powersum (i.e. a sum of powers with complex 
   exponents) and its transformation to the corresponding operational tree */

public class ComplexPowerSum {

  /** contains the coefficient for a constructed logarithm (which may happen,
      when integrating z^-1) */
  protected Complex logCoeff;

  /** a flag indicating, that an integration of z^-1 has happened, this flag
      must be checked, to test if further operations (other than eval())
      are allowed. It is set to false, if  */
  protected boolean valid = true;

  /** An Array of the coefficients */
  protected Complex[] coefficients;
 
  /** the corresponding powers */
  protected Complex[] powers;
  
  /** Nullpowersums (i.e. coefficients.length == 0) are allowed, but not
      a nullpointer argument! */

  public ComplexPowerSum(Complex[] coefficients, Complex[] powers){

    this.coefficients = new Complex[coefficients.length];
    this.powers = new Complex[powers.length];
    System.arraycopy(coefficients, 0 , this.coefficients, 0, 
		     coefficients.length);
    System.arraycopy(powers, 0 , this.powers, 0, powers.length);
  }  

  
  /** convert a polynomial to a ComplexPowerSum */

  public ComplexPowerSum(Polynom p){
    int degree = p.degree();
    Complex[] newCoefficients = new Complex[degree];
    Complex[] newPowers = new Complex[degree];

    int count = 0;
    for(int i = 0; i < degree; i++)
      // use only the non-zero coefficients
      if(p.getCoefficient(i) != 0.0d){
	newPowers[count] = new Complex(i,0);
	newCoefficients[count] =  new Complex(p.getCoefficient(i),0);
	count++;
      }
    
    // copy the values into the attributes
    this.coefficients = new Complex[count];  
    this.powers = new Complex[count];  
    System.arraycopy(newCoefficients, 0 , this.coefficients, 0, 
		     count);
    System.arraycopy(newPowers, 0 , this.powers, 0, count);
  }  



  public ComplexPowerSum(){
  }


  /** return a coefficient of a single indexed monom */
  public Complex getCoefficient(int i) throws InvalidException{
    if(!valid) throw new InvalidException();
    return coefficients[i];
  }
 
  /** returns the power of a single indexed monom */
  public Complex getPower(int i) throws InvalidException{
    if(!valid) throw new InvalidException();
    return powers[i];
  }

  /** sets a single specified coefficient */
  public void setCoefficient(int i, Complex value) throws InvalidException{
    if(!valid) throw new InvalidException();
    coefficients[i] = value;
  }

  /** sets a single specified coefficient with the according power */
  public void setCoefficient(int i, Complex value, Complex power) 
    throws InvalidException {
    if(!valid) throw new InvalidException();
    coefficients[i] = value;
    powers[i] = power;
  }

  /** returns all coefficients */
  public Complex[] getCoefficients() throws InvalidException{
    if(!valid) throw new InvalidException();
    return coefficients;
  }

  /** returns all powers */
  public Complex[] getPowers() throws InvalidException{
    if(!valid) throw new InvalidException();
    return powers;
  }

  /** multiplication with a number */
  public void multiply(Complex d) throws InvalidException{
    if(!valid) throw new InvalidException();
    for(int i = 0; i < coefficients.length; i++)
      coefficients[i].mult(d);
  }


  /** multiplication with another powersum */
  public void multiply(ComplexPowerSum p) throws InvalidException{
    //System.out.println("multiplying "+this+" and "+p);
    if(!valid) throw new InvalidException();
    Complex[] pCoeffs = p.getCoefficients();
    Complex[] pPowers = p.getPowers();

    // the length of the new powersum is maximally the product of the lengths 
    // of both factors. It is less, if there are common powers in both factors
    int maxlength = pCoeffs.length * coefficients.length;

    // we use a hashtable for storing the different coefficients with the
    // associated power as key first, so coefficients for equal powers are 
    // stored with the same key
    Hashtable powerHash = new Hashtable(maxlength);

    // the coefficients of each pair of single monoms are multiplied, powers 
    // are added
    for(int i = 0; i < coefficients.length; i++)
      for(int j = 0; j < pCoeffs.length; j++){

	// check if the hashtable is empty for this key
	Complex sumOfPower = Complex.add(powers[i], (pPowers[j]));
	Complex productOfCoeffs = Complex.mult(coefficients[i], pCoeffs[j]);
	if(!powerHash.containsKey(sumOfPower)){

	  // simply put in new value
	  powerHash.put(sumOfPower, productOfCoeffs);
	} else {
	  
	  // we need to add the already existing value to the one we want
	  // to put in
	  Complex existingValue = (Complex) powerHash.get(sumOfPower);

	  powerHash.put(sumOfPower, 
			Complex.add(productOfCoeffs,
				    existingValue));	  
	}
       
      }
	
    // now we put the result in arrays, which then replace the original 
    // arrays (powers & coeffs)  
    Enumeration coeffEnum = powerHash.elements();
    Enumeration powerEnum = powerHash.keys();
    Complex[] newCoeffs = new Complex[powerHash.size()];
    Complex[] newPowers = new Complex[powerHash.size()];
    int i = 0;
  
    while(coeffEnum.hasMoreElements()){
      Complex d_coeff = (Complex) coeffEnum.nextElement();
      Complex d_power = (Complex) powerEnum.nextElement();
      
      // we count only the values, where coeff != 0
      if(!d_coeff.equals(Complex.ZERO)){
	newCoeffs[i] = d_coeff;
	newPowers[i] = d_power;
	i++;
      }
    }

    // copy the arrays into the original ones, using i as size (thus
    // eliminating the empty space for zero-coefficients      
    powers = new Complex[i];
    coefficients = new Complex[i];
    System.arraycopy(newPowers, 0, powers, 0, i);
    System.arraycopy(newCoeffs, 0, coefficients, 0, i);

  }



  /** evaluates the powersum for a certain Complex number */
  public Complex eval(Complex z){
    Complex result = new Complex(0,0);
    for(int i = 0; i < coefficients.length; i++){
      //check for singularities: if found, we need to use the logarithm
      if(Double.isNaN(coefficients[i].real) ||
	 Double.isNaN(coefficients[i].imag))
	// use the value stored in logCoeff
	result.add(Complex.mult(logCoeff, 
				Complex.ln(z)));
      else 
					  
	result.add(Complex.mult(coefficients[i], 
				Complex.pow(z,powers[i])));
    }
    return result;
  }


  /** sorts the arrays in ascending order of powers */
  public void sort(){
    
    //System.out.println("\n\n old is: "+toString()+", length is "+powers.length);

    // we use a hashtable for storing the different coefficients with the
    // associated power as key
    // we do this, so we can make use of the java internal sorting 
    // algorithms buried in java.util.Arrays

    Hashtable powerHash = new Hashtable();
    
    for(int i = 0; i < coefficients.length; i++) {
      // put the corresponding coefficient in the hash with the power as key
      powerHash.put(new Complex(powers[i]), new Complex(coefficients[i]));
    }
    
    // now sort the powers in ascending order
    Arrays.sort(powers);

    Complex[] newCoeffs = new Complex[powers.length];

    // re-connect the coefficients to the corresponding powers by descending
    // through the array in reverse order
    for(int i = 0; i < powers.length; i++){
      Complex coefficient = (Complex) powerHash.get(new Complex(powers[i]));
      newCoeffs[i] = coefficient;
    }
    
    coefficients = newCoeffs;

  }


  /** integrates the powersum numerically, using the arguments as range */
  public void integrate(Complex constant) throws InvalidException{
    if(!valid) throw new InvalidException();
    //System.out.println("integrating :"+toString());

    Complex[] newCoefficients = new Complex[coefficients.length + 1];
    Complex[] newPowers = new Complex[coefficients.length + 1];

    newCoefficients[0] = constant;
    newPowers[0] = new Complex(0,0);

    for(int i = 0; i < coefficients.length; i++){
      if(powers[i].equals(new Complex(-1,0))){	
	valid = false;	
	logCoeff = new Complex(coefficients[i]);
	//System.out.println("logCoeff is: "+logCoeff);
      } 
      newCoefficients[i+1] = Complex.div(coefficients[i], 
					 Complex.add(powers[i], 
						     Complex.ONE));
      newPowers[i+1] = Complex.add(powers[i], Complex.ONE);
    }

    coefficients = newCoefficients;
    powers = newPowers;
    //System.out.println("result is :"+toString());

  }

  /** for pretty printing: output is of form: 
      <pre>
      X^6 - 3.0X^5 + 2.5X^4 - 0.49999999999999956X^2 + 0.023809523809523885
      </pre>
  **/
  public String toString(){
    String retVal = "";
    for(int i = powers.length - 1 ; i >= 0;  i--){
      retVal += " +" + " " + coefficients[i] + 
	(!powers[i].equals(Complex.ZERO) ? "*X^" + powers[i] : "");
    }
	  
    // eliminate first " + "
    retVal = retVal.substring(3);
    return retVal;
  }


  /** builds and returns the corresponding operational tree for further
      calculations */
  public BaseCalc getOperationTree(){

    // if powersum is the nullpowersum, return constant zero
    if(coefficients.length == 0)
      return new NumeralCalc(new Complex(0,0));

    // handle special case of constant powersum
    if(coefficients.length == 1)
      return new NumeralCalc(coefficients[0]);

    // from now on, the the "normal" cases are handled

    AddCalc add = new AddCalc();
    // this is the node of the previous add-Operation, initialized as 0;
    BaseCalc previous = new NumeralCalc(new Complex(0,0));

    try{ 
      // all coefficients except for the additional constant have the same
      // structure
      for(int i = coefficients.length - 1 ; i > 0;  i--){
	
	// construct the operational tree for a single power: coefficients[i]*X^powers[i]
	BaseCalc[] power = {new VariableCalc(), new NumeralCalc(powers[i])};
	
	PowerCalc p = new PowerCalc();
	p.insertChildren(power);
	
	BaseCalc[] term = {new NumeralCalc(coefficients[i]), p};
	
	MultiplyCalc m = new MultiplyCalc();			   
	m.insertChildren(term);
	
	// connect this node to the previuos add-operation
	BaseCalc[] args = {previous,m};
	add.insertChildren(args);
	
	// store the old operational tree in previous and create a new 
	// add-operation
	previous = add;
	add = new AddCalc();
      }
      // store constant and connect it to the add-operation 
      NumeralCalc constant = new NumeralCalc(coefficients[0]);
      BaseCalc[] args = {previous, constant};
      add.insertChildren(args);
    } catch (IllegalArgException e) { System.out.println(e);}
    
    return add;
  }

  /** for testing purposes */
  public static void main(String[] args){
    try {

      Complex[] coeffs = {new Complex(1,1)};
      Complex[] powers = { new Complex(-1,0)};
      ComplexPowerSum p = new ComplexPowerSum(coeffs, powers);
      ComplexPowerSum q = new ComplexPowerSum(coeffs, powers);
      System.out.println("p := "+p.toString());
      System.out.println(q.toString());
      /* System.out.println("q := p^2 is: ");    
	 q.multiply(p); */
      q.integrate(new Complex(0,0));
      System.out.println("q := int(p,x) is: ");
      System.out.println(q.toString());
      System.out.println(Complex.subtract(q.eval(new Complex(2,0)),
					  q.eval(new Complex(1,0))).toString());
      //System.out.println("OperationTree is :" + q.getOperationTree().toString());
      //System.out.println(p.eval(new Complex(1,0)).toString());
      /*
	double[] coeffs2 = { 1.0d, 3.0d , 1.0d, 4.0d};
	double[] coeffs3 = { 2.0d, 1.0d , 1.0d};
	ComplexPowerSum p2 = new ComplexPowerSum(new Polynom(coeffs2));
	ComplexPowerSum q2 = new ComplexPowerSum(new Polynom(coeffs3));
	System.out.println("\n(" + p2.toString() +") * ("+ q2.toString() +") is :");
	p2.multiply(q2);
	System.out.println(p2.toString()); 
      */
    } catch(InvalidException e){ e.printStackTrace();}
  }
}
