package calc;

import java.util.*;


/** This class takes a string representing an arithmetic expression, a
    pair of ranges and performs the calculations for a twodimensional grid 
    of elements in this range by feeding the operational tree (produced
    by <code>ExprParser</code>) with the apropriate values 

    @see ExprParser */

public class DoCalc {
    
  public final static int DEFAULT_STEPS = 30;

  /** used to store the representations of variables in the operational 
      tree */

  private Vector variableCalcs = null;

  /** the root node of the operational tree, returned by 
      <code>ExprParse</code> */

  private BaseCalc opRoot = null;

  /** the Range on the real-axis (x) */

  private double[] realRange = {-10,10};

  /** the Range on the imaginary-axis (y) */

  private double[] imagRange = {-10,10};

  /** the number of steps in the grid */

  private int steps = DEFAULT_STEPS;


  /** the twodimensional ((steps+1) x (steps+1)) array of Complex Numbers 
      containing the calculated value for the specified grid */

  private Complex[][] values = new Complex[DEFAULT_STEPS+1][DEFAULT_STEPS+1];



  /** preparses the string, handles it to <code>ExprParser</code>,
      stores the operational tree and	keeps record of the Modules 
      representing the variable z for later insertion (i.e fills
      <code>variableCalc</code> with elements of Type 
      <code>VariableCalc</code> */
    
  public DoCalc(String expr){
	
    String preparsedExpr = preparse(expr);

    try {
      ExprParser parse = new ExprParser(preparsedExpr);
      opRoot = parse.getOperationTree();
      variableCalcs = parse.getVariableCalcs();
    } catch(SyntaxErrorException e) { e.printStackTrace(); }

	
  }


  /** lower cases input and does the following replacements:
      <pre>
      exp -> e
      sin -> s
      cos -> c
      zeta -> 
      re -> x
      im -> m
      </pre>
  */

  private String preparse(String expr){

    expr = expr.toLowerCase();
    expr = replace(expr,"zeta","");
    expr = replace(expr,"exp","e");
    expr = replace(expr,"sin","s");
    expr = replace(expr,"cos","c");
    expr = replace(expr,"re","r");
    expr = replace(expr,"im","m");
	
    System.out.println("preparsed to "+expr);
    return expr;
  }

  /** does the dirty work for preparse() (why is there no regexp-class?!) */

  private String replace(String expr, String what, String by){
    int i = 0;
    i = expr.indexOf(what,i);
    while(expr.regionMatches(i, what, 0, what.length())) {
      System.out.println(i);
    
      expr = expr.substring(0,i) + by 
	+ expr.substring(i + what.length());
      System.out.println(expr);
      i = expr.indexOf(what,i);
    
    }
    return expr;
  }

  /** set the range of the grid to be calculated 
      (default is (-10,10) in each direction) */

  public void setRange(double[] realRange, double[] imagRange){
    this.realRange = realRange;
    this.imagRange = imagRange;
  }

  /** set the number of steps in the grid (default is 30, so 30x30 values
      are calculated) */

  public void setSteps(int steps){ 
    this.steps = steps;
    values = new Complex[steps+1][steps+1];
  }

  /** get the Grid of Values for this Calculation */

  public Complex[][] getValues(){
    return values;
  }
	
  public String toString(){
    String retVal = "[";
    for(int i = 0; i <= steps; i++)
      for(int j = 0; j <= steps; j++)
	retVal += " "+values[i][j].toString();
	    
    return retVal;
  }


  /** calculates a result for a certain number */
  public Complex calculateFor(Complex z){
    opRoot.clear();
    // insert the argument def into the variable-holders
    for(int k = 0; k < variableCalcs.size(); k++)
      ((VariableCalc) variableCalcs.elementAt(k)).setValue(z);
    return new Complex(opRoot.result());

  }
  

  /** iterates over the grid and produces the results, storing them for
      retrieval with <code>getValues()</code> */

  public void calculate(){

    double width = realRange[1] - realRange[0];
    double height = imagRange[1] - imagRange[0];

    // calculate an equidistant grid within the range and
    // evaluate the operational tree for each value of z in the grid
	
    for(int i = 0; i <= steps; i++)
      for(int j = 0; j <= steps; j++){
	Complex def = new Complex(
				  realRange[0] + (i * width)  / steps,
				  imagRange[0] + (j * height) / steps );
		
	// insert the argument def into the variable-holders
	for(int k = 0; k < variableCalcs.size(); k++){
	  ((VariableCalc) variableCalcs.elementAt(k)).setValue(def);
		    
	  //		    System.out.println("Value is :"
	  //  +((VariableCalc)variableCalcs.elementAt(k)).result());
	}
	// calculate the result of the whole operation and store 
	// it in the grid
	opRoot.clear();
	values[i][j] = opRoot.result();
      }
  }
	

  /** called by user */

  public static void main(String[] args){
    DoCalc Calc = new DoCalc(args[0]);
    System.out.println("OPTree is : "+Calc.opRoot.toString());
    Calc.calculate();
    System.out.println(Calc.toString());
  }
}


	
