package calc;

import java.util.*;

/* This class represents a polynomial and its transformation to the 
   corresponding operational tree */

public class Polynom {

  /** An Array of the coefficients, in ascending power (i.e. beginning with
      the constant summand) */
  protected double[] coefficients;

  /** the default constructor creates the nullpolynomial */
  public Polynom(){
    this(new double[0]);
  }

  /** Nullpolynomials (i.e. coefficients.length == 0) are allowed, but not
      a nullpointer argument! */
  public Polynom(double[] coefficients){

    this.coefficients = new double[coefficients.length];
    System.arraycopy(coefficients, 0 , this.coefficients, 0, 
		     coefficients.length);
  }  

  /** Copy-Constructor */
  public Polynom(Polynom p){
    this(p.getCoefficients());
  }

  public int degree(){
    return coefficients.length;
  }

  public double getCoefficient(int i){
    return coefficients[i];
  }

  public void setCoefficient(int i, double value){
    coefficients[i] = value;
  }

  public double[] getCoefficients(){
    return coefficients;
  }

  /** adds two polynomials */
  public void add(Polynom p){
    double[] newCoefficients = new double[Math.max(p.degree(), degree())];

    // add for each single coefficients
    for(int i = 0; i < newCoefficients.length; i++){

      // check if any of the polynomials is of lesser size than i
      // if not, add the coefficient to the resulting coefficient
      if(i < degree())
	newCoefficients[i] += getCoefficient(i);

      if(i < p.degree())
	newCoefficients[i] += p.getCoefficient(i);
    }
   
    coefficients = newCoefficients;
  }


  /** multiplication with a number */
  public void multiply(double d){
    for(int i = 0; i < degree(); i++)
      coefficients[i] *= d;
  }


  /** multiplication with another polynomial */
  public void multiply(Polynom p){
    double[] pCoeffs = p.getCoefficients();

    // grad(pNew) = grad(this) + grad(p) 
    double[] newCoeffs = new double[p.degree()+degree()];

    for(int i = 0; i < coefficients.length; i++)
      for(int j = 0; j < pCoeffs.length; j++)
	newCoeffs[i+j] += coefficients[i] * pCoeffs[j];
    coefficients = newCoeffs;
  }    

  /** performs the operation f(x) -> f(x+k) */

  public void translate(double k){
    
    // we perform this operation by successively adding k to all monomials, so
    // that a*x^b -> a*(x+k)^b and then we create by multiplication and
    // summation the new polynomial

    // the result, initialize as nullpolynomial to which the rest is added
    Polynom result = new Polynom();

    for(int i = 0; i < coefficients.length; i++){

      // the first factor
      double[] one = {1};
      Polynom factor = new Polynom(one);

      // create a linear polynomial of the form x+k 
      double[] coeffs = {k, 1};
      Polynom translation = new Polynom(coeffs);

      // the raising to power i occurs by successive multiplication
      for(int j = 0; j < i; j++)
	factor.multiply(translation);
      
      // multiply the monomial with the corresponding coefficient
      factor.multiply(coefficients[i]);

      // now add the monomial to the result, creating the polynomial
      result.add(factor);
    }     
    coefficients = result.getCoefficients();
    //System.out.println("Result is: "+this);
  }


  /** evaluates the polynomial for a certain complex number */
  public Complex eval(Complex z){
    Complex result = new Complex(0,0);
    for(int i = 0; i < coefficients.length; i++)
      result = Complex.add(Complex.mult(new Complex(coefficients[i], 0),
					Complex.pow(z, new Complex(i,0))),
			   result);
    
    return result;
  }

  /** integrates the polynomial analytically, using the argument as integration
      constant */
  public void integrate(double constant){

    double[] newCoefficients = new double[degree() + 1];

    newCoefficients[0] = constant;

    for(int i = 0; i < degree(); i++)
      newCoefficients[i+1] = coefficients[i] / (double)(i+1);

    coefficients = newCoefficients;
  }

  /** for pretty printing: output is of form: 
      <pre>
      X^6 - 3.0X^5 + 2.5X^4 - 0.49999999999999956X^2 + 0.023809523809523885
      </pre>
  **/
  public String toString(){
    String retVal = "";
    for(int i = coefficients.length - 1 ; i >= 0;  i--)
      if(coefficients[i] != 0){
	if(coefficients[i] > 0)
	  // display "X^n" instead of "1X^n"
	  if(coefficients[i] == 1.0d && i != 0)	  
	    retVal += " +" + " " + (i != 0 ? "X^" + i : "");
	  else
	    retVal += " +" + " " + coefficients[i] + (i != 0 ? "X^" + i : "");
	else
	  // display "X^n" instead of "1 X^n"
	  if(coefficients[i] == -1.0d && i != 0)	  
	    retVal += " -" + " " + (i != 0 ? "X^" + i : "");
	  else
	    retVal += " -" + " " + (-coefficients[i]) + (i != 0 ? "X^" + i : "");

      }
	  
    // eliminate first " + "
    retVal = retVal.substring(3);
    return retVal;
  }


  /** builds and returns the corresponding operational tree for further
      calculations */
  public BaseCalc getOperationTree(){

    // if polynomial is the nullpolynomial, return constant zero
    if(coefficients.length == 0)
      return new NumeralCalc(new Complex(0,0));

    // handle special case of constant polynomial
    if(coefficients.length == 1)
      return new NumeralCalc(new Complex(coefficients[0],0));

    // from now on, the the "normal" cases are handled

    AddCalc add = new AddCalc();
    // this is the node of the previous add-Operation, initialized as 0;
    BaseCalc previous = new NumeralCalc(new Complex(0,0));

    try{ 
      // all coefficients except for the additional constant have the same
      // structure
      for(int i = coefficients.length - 1 ; i > 0;  i--){
	
	// construct the operational tree for a single power: coefficients[i]*X^i
	BaseCalc[] power = {new VariableCalc(), new NumeralCalc(new Complex(i,0))};
	
	PowerCalc p = new PowerCalc();
	p.insertChildren(power);
	
	BaseCalc[] term = {new NumeralCalc(new Complex(coefficients[i],0)), p};
	
	MultiplyCalc m = new MultiplyCalc();			   
	m.insertChildren(term);
	
	// connect this node to the previuos add-operation
	BaseCalc[] args = {previous,m};
	add.insertChildren(args);
	
	// store the old operational tree in previous and create a new 
	// add-operation
	previous = add;
	add = new AddCalc();
      }
      // store constant and connect it to the add-operation 
      NumeralCalc constant = new NumeralCalc(new Complex(coefficients[0],0));
      BaseCalc[] args = {previous, constant};
      add.insertChildren(args);
    } catch (IllegalArgException e) { System.out.println(e);}
    
    return add;
  }

  /** for testing purposes */
  public static void main(String[] args){
    double[] d = { 1.0d, 1.0d , 2.0d};
    Polynom p = new Polynom(d);
    Polynom q = new Polynom(d);
    System.out.println(q.toString());
    p.integrate(0.0);
    System.out.println("p := int(q,x) is: "+p.toString());
    //System.out.println(p.getOperationTree().toString());
    //System.out.println(p.eval(new Complex(1,0)).toString());
    q.translate(2.0d);
    System.out.println("q(x+2) is :" + q.toString());
    
  }
}
