package show;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;
import calc.Complex;

/** shows the values of a complex function (handed over in a Complex[][] array)
    as a map with Lines of constant Altitude */

public class AltitudeDisplay extends Field2DDisplay {

  /** This Array contains the sorted values (needed for partitioning into
      colorlevels) */
  protected double[] sortedValues;

  /** This Array contains the sorted mapping from colorindex to 
      corresponding (double) value. Each of the 255 colorlevels (ranging 
      from black to white) is associated with a (double) value. When painting 
      occurs, for each value v the array is iterated, stopping at the index i,
      if( valuesForColor[i] > v ).  */
  protected double[] valuesForColor = new double[255];

  /** canvas is initialized as AltitudeCanvas in getCanvas(), when called the
   first time */
  protected Canvas canvas = null;

  public Canvas getCanvas(){
    if(canvas == null)
      canvas = new AltitudeCanvas();
    
    return canvas;
  }

  /** copies the complex values from the original ones and stores them
      also as absolute */
  public void setValues(Complex[][] inValues){
    n = inValues.length;
    m = inValues[0].length;

    values = new Complex[n][m];
    sortedValues = new double[n*m];
    for(int i = 0; i < n; i++)
      for(int j = 0; j < m; j++){

	values[i][j] = new Complex(inValues[i][j]);

	// first simply put in the values, sorting occurs at the end
	sortedValues[i*m + j] = inValues[i][j].square();
      }

    if(sortedValues.length == 0) return;

    //System.out.println("sorting...");
    Arrays.sort(sortedValues);

    // balance the valuesForColor Array, so that every colorrange
    // (i.e. every valuesForColor[i]) has an equal amount of 
    // values for which valuesForColor[i] < x <= valuesForColor[i+1]
    // this means we simply divide sortedValues into 255 partitions
    // inserting the left interval-border of the i-th partition into 
    // valuesForColor[i]

    for (int i = 0; i < 255; i++)
      valuesForColor[i] = sortedValues[(n*m)/255 * i]; 
  
  }

  /** this inner class represents the canvas on which everything is
      drawn. We use physical coordinates (depending on the size of the 
      component) and logical coordinates (depending on the given definition 
      range) which are transformed by methods of the outer class */
  public class AltitudeCanvas extends Canvas {
    public void paint(Graphics g){
      if(n == 0 || m == 0)
	return;

      determineCellSize();
      
      // we use the Java2D-API
      Graphics2D g2 = (Graphics2D) g;
      //g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
      //RenderingHints.VALUE_ANTIALIAS_ON); -- too slow!

      // set transform by calling Field2DDisplay.determineTransform()
      determineTransform(getSize().width,
			 getSize().height - showCoords.getSize().height);

      g2.setTransform(transform);	
		    
      // iterate through the grid of values
      for(int i = 0; i < n; i++)
	for(int j = 0; j < m; j++){
		
	  // iterate through valuesForColor[], stop at the index where 
	  // value is greater than the colorlevel-value
	  int k = 0;

	  while(k < 255 && values[i][j].square() >= valuesForColor[k]  )
	    k++;
	  // paint the cell green if it is inValid, otherwise apply the value k
	  if(values[i][j].isValid())
	    g2.setColor(new Color(k, k, k));
	  else
	    g2.setColor(new Color(0, 255, 0));	    
	  g2.fill(new Rectangle2D.Double(realRange[0] + i*cellWidth, 
					 imagRange[0] + j*cellHeight, 
					 cellWidth, cellHeight));

	}

      // draw coordinates

      g2.setColor(Color.red);

      g2.setStroke(new BasicStroke((float) (2 * imagDif / getSize().height),
				   BasicStroke.CAP_BUTT, 
				   BasicStroke.JOIN_MITER));
      g2.draw(new Line2D.Double(realRange[0], 0, realRange[1], 0));
      g2.setStroke(new BasicStroke((float) (2 * realDif / getSize().width),
				   BasicStroke.CAP_BUTT, 
				   BasicStroke.JOIN_MITER));
      g2.draw(new Line2D.Double(0, imagRange[0], 0, imagRange[1]));
    }
      
  }



  public AltitudeDisplay(){

    // initialize, so that paint() doesn't complain
    Complex[][] nullValues = {{}};
    setValues(nullValues);
	  
  }

    
}
	
