package show;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import calc.Complex;

/** the base class for a 2-dimensional view of complex functions it implements
    the functionality needed for transforming the value-canvas to the chosen
    size. This is done by using the Java2D-API */

public abstract class Field2DDisplay extends ComplexDisplay {

  /** needed here for output only */
  protected Complex[][] values = {{new Complex(0,0)}};

  /** the canvas is stored in this variable */
  private Canvas c = null;

  /** the values array has a height of m and a width of n */
  int n = 0, m = 0;

  /** the range in real direction, needs to be stored for transformation */
  protected double[] realRange = {-10.0d,10.0d};

  /** the range in imaginary direction, needs to be stored for 
      transformation */
  protected double[] imagRange = {-10.0d,10.0d};

  /** a label showing the complex coordinates of the mouse*/
  protected Label showCoords = new Label("Coords");

  /** realRange[1] - realRange[0], needed for transformation */
  protected double realDif = 0;

  /** imagRange[1] - imagRange[0], needed for transformation */
  protected double imagDif = 0;

  /** the width of a single cell in (logical) pixels */
  protected double cellWidth = 1; 

  /** the height of a single cell in (logical) pixels */
  protected double cellHeight = 1; 

  /** the last chosen point (by MouseCoordinateReader) */
  protected Point circleCenter = new Point(0,0);

  /** the last coordinate used in MouseCoordinateReader.mouseDragged (for 
      reversing)*/
  protected Point circleOuter = new Point(0,0);
  
  /** the affine transformation applied to the Complex-Data for projecting
      it on the screen, its inverse is used by the MouseCoordinateReader to
      calculate the Complex coordinates from physical screen-coordinates */
  protected AffineTransform transform = new AffineTransform();

  /** store the mouselistener used by this class, so that the subclass can
      remove it by reference */
  protected MouseCoordinateReader mouseCoordinateReader = new MouseCoordinateReader();

  /** stores the calculator frontend to compute function values independent
      of the value-grid */
  protected calc.Calculator calculator;

  /** sets the calculator frontend */
  public void setCalculator(calc.Calculator calculator){
    this.calculator = calculator;
  }

  /** returns a Complex number with points which have been transformed
      from integer-physical graphics coordinates to double logical space
      coordinates  using the inverse Transformation of the one used in
      paint() */
  public Complex transformXY(int x, int y){
    double[] in = {x,y};
    double[] out = new double[2];
    try {
      transform.createInverse().transform(in,0,out,0,1);
    } catch(Exception ex){ex.printStackTrace();}
    return new Complex(out[0], out[1]);
  }

  /** return the computed function value for a given complex number using 
      the grid of values (warning: is inaccurate compared to 
      calculator.calculateForCurrentFunction()!) */
  public Complex getValue(Complex c){

    int[] indices = getIndices(c);
    return new Complex(values[indices[0]][indices[1]]);
  }

  /** return the index pair for a given complex number using 
      the ranges and gridsize */
  public int[] getIndices(Complex c){

    int i = (int) (((c.real - realRange[0]) * n) / realDif);
    int j = (int) (((c.imag - imagRange[0]) * m) / imagDif);
    if(i >= n) i = n - 1;
    if(j >= n) j = m - 1;
    if(i < 0) i = 0;
    if(j < 0) j = 0;
    int[] result = {i,j};
    return result;
  }


  /** sets the cellwidth and -height, determined by realDif and imagDif */
  public void determineCellSize(){
    cellWidth = (double) realDif / n;
    cellHeight = (double) imagDif / m;
  }

  /** sets the affine transformation, determined by window-size and readDif,
      imagDif */
  public void determineTransform(int width, int height){
    // scale the Image to the appropriate size
    double factorx = ((double) width) / realDif;
    double factory = ((double) height) / imagDif;
    
    transform = new AffineTransform(factorx, 0,
				    0, - factory,
				    - realRange[0] * factorx, 
				    imagRange[1] * factory);
  }
  
  /** sets the ranges for both real and imaginary direction */
  public void setRange(double[] realRange, double[] imagRange){
    this.realRange[0] = realRange[0];
    this.realRange[1] = realRange[1];
    this.imagRange[0] = imagRange[0];
    this.imagRange[1] = imagRange[1];
    realDif = realRange[1] - realRange[0];
    imagDif = imagRange[1] - imagRange[0];
  }

  /** an inner class, which reads the coordinates of the mouse and transforms
      them to complex coordinates. Also used for determining the gradient
      within a given radius */
  
  public class MouseCoordinateReader extends MouseAdapter implements MouseMotionListener {
    
    public void mouseMoved(MouseEvent e){

      // read out mouse-coordinates (modified by offsets) and apply inverse 
      // transformation on them
      Complex c = transformXY(e.getX(), e.getY());

      // determine function value for c
      double f_x = calculator.calculateForCurrentFunction(c).real;
      double f_y = calculator.calculateForCurrentFunction(c).imag;

      // if numbers are of form xxxE-xx (i.e. too low), they must be set to 
      // zero for displaying purposes
      if(Math.abs(c.real) < 0.001d)
	c.real = 0.0d;
      if(Math.abs(c.imag) < 0.001d)
	c.imag = 0.0d;
      if(Math.abs(f_x) < 0.001d)
	f_x = 0.0d;
      if(Math.abs(f_y) < 0.001d)
	f_y = 0.0d;	

      // here we also have to add some Code here for numbers of form xxxExx 
      // (too high)

      // beautify all results by cropping the numbers to 3 digits after '.'
      String outx = Double.toString(c.real);
      String outy = Double.toString(c.imag);
      String outf_x = Double.toString(f_x);
      String outf_y = Double.toString(f_y);
      if(outx.indexOf(".") + 3 < outx.length()) 
	outx = outx.substring(0, outx.indexOf(".") + 4)+(outx.indexOf("E") > -1 ? outx.substring(outx.indexOf("E")-1,outx.length()) : "");      
      
      if(outy.indexOf(".") + 3 < outy.length()) 
	outy = outy.substring(0, outy.indexOf(".") + 4)+(outy.indexOf("E") > -1 ? outy.substring(outy.indexOf("E")-1,outy.length()) : "");      
      
      if(outf_x.indexOf(".") + 3 < outf_x.length()) 
	outf_x = outf_x.substring(0, outf_x.indexOf(".") + 4)+(outf_x.indexOf("E") > -1 ? outf_x.substring(outf_x.indexOf("E")-1,outf_x.length()) : "");      
      
      if(outf_y.indexOf(".") + 3 < outf_y.length()) 
	outf_y = outf_y.substring(0, outf_y.indexOf(".") + 4)+(outf_y.indexOf("E") > -1 ? outf_y.substring(outf_y.indexOf("E")-1,outf_y.length()) : "");      
      
      showCoords.setText("z = ("+outx+","+outy+"i);      f(z) = ("
			 +outf_x+","+outf_y+"i)");
    }
    
      /** this method reads out the mouse coordinates, which are used by
	  mouseDragged to put a circle around it, in order to calculate the 
	  gradient
      */

      public void mousePressed(MouseEvent e){
      
	int x = e.getX();
	int y = e.getY();
      
	// reset values: center is new point, outer is initially in the center
	circleCenter = new Point(x,y);
	circleOuter = circleCenter;

      }


      /** paints a circle with radius of the current distance from circleCenter
	  (set in mousePressed()) */
      public void mouseDragged(MouseEvent e){
	Point newOuter = new Point(e.getX(),e.getY());
	Graphics g = c.getGraphics();
	g.setXORMode(Color.blue);

	// determine old circle by the radius of the minimum distance from 
	// circleCenter
	int radius = (int) circleOuter.distance(circleCenter);
      
	// erase old circle 
	g.drawOval(circleCenter.x - radius, circleCenter.y - radius, 
		   2*radius, 2*radius);
      
	// change to new circle-outline and determine new radius
	circleOuter = newOuter;
	radius = (int) circleOuter.distance(circleCenter);

	g.drawOval(circleCenter.x - radius, circleCenter.y - radius, 
		   2*radius, 2*radius);
      }      

      /** calculates the gradient for the circle determined by mousePressed
	  and mouseDragged */

      public void mouseReleased(MouseEvent e){
      
	// first erase the last drawn circle
	int radius = (int) circleOuter.distance(circleCenter);
	Graphics g = c.getGraphics();
	g.setXORMode(Color.blue);
	g.drawOval(circleCenter.x - radius, circleCenter.y - radius, 
		   2*radius, 2*radius);

	// from now on we use logical coordinates, thus we need to switch
	// to graphics2d functionality
	Graphics2D g2 = (Graphics2D) g;

	// set line width to two physical pixel (in logical coordinates!)
	g2.setStroke(new BasicStroke((float)(2 * realDif / c.getSize().width), 
				     BasicStroke.CAP_BUTT, 
				     BasicStroke.JOIN_MITER));
      
	// keep the absolute function value for the center of the circle 
	Complex center = transformXY(circleCenter.x, circleCenter.y);
	Complex f_center = calculator.calculateForCurrentFunction(center);

	// a and b are the half axes of the logical space ellipse to which
	// the physical circle is transformed
	double a = Complex.subtract(transformXY(0,0),
				    transformXY(radius,0)).abs();
	double b = Complex.subtract(transformXY(0,0),
				    transformXY(0,radius)).abs();

	//System.out.println("physical radius is: "+radius);
	//System.out.println("Drawing ellipse around "+center+"; with a,b = "
	// + a + "," + b);

	// now walk through a circle with the chosen radius (using sin and cos) 
	// and determine for each point the function value

	Complex max_c = new Complex(0, 0);
	double max_f_diff = Double.NEGATIVE_INFINITY, x = 0, y = 0, f_diff = 0;

	// use logical coordinates
	g2.setTransform(transform);	
      
	// paint the gradient in red
	g.setPaintMode();
	g.setColor(Color.red);
      
	for(int degree = 1; degree <= 360; degree++){
	  double angle = (degree * Math.PI)/180;

	  // we use logical coordinates; keep aspect ratio (realdif/imagdif) in 
	  // mind!
	  x = (double) center.real + Math.cos(angle) * a;	  
	  y = (double) center.imag + Math.sin(angle) * b;


	  Complex c = new Complex(x, y);

	  // determine function values for the circle line and calculating
	  // the difference, thus building the finite version of the 
	  // gradient
	  double f_circle = calculator.calculateForCurrentFunction(c).abs();
	  f_diff = f_center.abs() - f_circle;
	  //g2.fill(new Rectangle2D.Double(c.real, c.imag, cellWidth, cellHeight));
	  //System.out.println("value at: "+c+"; f_circle is: "+f_circle);

	  // if we have a new maximum, replace the old one with it	
	  if(f_diff >= max_f_diff){
	    //System.out.println("New maximum at: "+c+"; f_circle is: "+f_circle);
	    max_c = c;
	    max_f_diff = f_diff;
	  }
	}
	

	g2.setColor(Color.green);
	Complex cc = transformXY(circleCenter.x, circleCenter.y);
	g2.draw(new Line2D.Double(cc.real, cc.imag, max_c.real, max_c.imag));
      
      }      
    }

    /* canvas is handled by child class. In order to get to it from this class,
       this method must be implemented by child */
    public abstract Canvas getCanvas();

    public Field2DDisplay(){
    
      // get Canvas from subclass, initializing it by this
      c = getCanvas();

      //to read out Complex Coordinates
      c.addMouseMotionListener(mouseCoordinateReader);
      c.addMouseListener(mouseCoordinateReader);
      setLayout(new BorderLayout());
      add(c, BorderLayout.CENTER);
      add(showCoords, BorderLayout.SOUTH);
    }
  
  }	
