package show;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import calc.Complex;

/** shows the values of a complex function (handed over in a Complex[][] array)
    by transforming a grid */

public class GridDisplay extends ComplexDisplay {

  /** needed here for output only */
  protected Complex[][] complexValues = {{new Complex(0,0)}};

  /** the canvas is stored in this variable */
  private Canvas c = null;

  /** the complexValues array has a height of m and a width of n */
  int n = 0, m = 0;

  /** the range in real direction, needs to be stored for transformation */
  protected double[] realRange = {-10.0d,10.0d};

  /** the range in imaginary direction, needs to be stored for 
      transformation */
  protected double[] imagRange = {-10.0d,10.0d};

  /** the range in view, initialized to imagRange */
  protected double[] viewRealRange = {-10.0d,10.0d};

  /** the range in view, initialized to imagRange */
  protected double[] viewImagRange = {-10.0d,10.0d};

  /** realRange[1] - realRange[0], needed for transformation */
  protected double realDif = 0;

  /** imagRange[1] - imagRange[0], needed for transformation */
  protected double imagDif = 0;

  /** the real diff in the view */
  protected double viewRealDif;

  /** the imaginary diff in the view */
  protected double viewImagDif;

  /** a label showing the function value of the coordinate of the mouse*/
  protected Label showCoords1 = new Label("Coords                             ");
  /** a label showing the coordinates of the mouse in complex numbers*/
  protected Label showCoords2 = new Label("Coords                             ");

  /** initialize canvas as GridCanvas in getCanvas(), when called the 
      first time */
  protected Canvas canvas = new GridCanvas();

  /** the affine transformation applied to the Complex-Data for projecting
      it on the screen, its inverse is used by the MouseCoordinateReader to
      calculate the Complex coordinates from physical screen-coordinates */
  protected AffineTransform transform = new AffineTransform();

  /** store the mouselistener used by this class, so that the subclass can
      remove it by reference */
  protected MouseCoordinateReader mouseCoordinateReader = new MouseCoordinateReader();

  /** the width of a single cell in (logical) pixels */
  protected double cellWidth = 1; 

  /** the height of a single cell in (logical) pixels */
  protected double cellHeight = 1; 

  /** sets the cellwidth and -height, determined by realDif and imagDif */
  public void determineCellSize(){
    cellWidth = (double) realDif / n;
    cellHeight = (double) imagDif / m;
  }

  /** the buttons for directing (zooming/panning) in the graph */
  DirPanel dirPanel = new DirPanel();

  /** a panel containing the arrow-buttons and doing the movement */
  private class DirPanel extends Panel {
    
    
    /** Constructor: initialize GUI, add Actionlisteners for Buttons */
    DirPanel(){

      Panel grid1 = new Panel();
      Panel grid2 = new Panel();
      Panel grid3 = new Panel();
      grid1.setLayout(new BorderLayout());
      grid2.setLayout(new BorderLayout());
      grid3.setLayout(new BorderLayout());
      Button up = new Button("up");
      Button down = new Button("down");
      Button left = new Button("left");
      Button right = new Button("right");
      Button zoomIn = new Button("zoom in");
      Button zoomOut = new Button("zoom out");

      up.addActionListener(new ActionListener(){
	public void actionPerformed(ActionEvent e){
	  up();
	}
      });
      down.addActionListener(new ActionListener(){
	public void actionPerformed(ActionEvent e){
	  down();
	}
      });
      left.addActionListener(new ActionListener(){
	public void actionPerformed(ActionEvent e){
	  left();
	}
      });
      right.addActionListener(new ActionListener(){
	public void actionPerformed(ActionEvent e){
	  right();
	}
      });
      zoomIn.addActionListener(new ActionListener(){
	public void actionPerformed(ActionEvent e){
	  zoomIn();
	}
      });
      zoomOut.addActionListener(new ActionListener(){
	public void actionPerformed(ActionEvent e){
	  zoomOut();
	}
      });

      grid1.add(up,BorderLayout.NORTH);
      grid1.add(down,BorderLayout.SOUTH);
      grid1.add(left,BorderLayout.WEST);
      grid1.add(right,BorderLayout.EAST);
      grid2.add(zoomIn,BorderLayout.NORTH);
      grid2.add(zoomOut,BorderLayout.SOUTH);
      grid3.add(showCoords1,BorderLayout.NORTH);
      grid3.add(showCoords2,BorderLayout.SOUTH);
      add(grid3);
      add(grid1);
      add(grid2);
    }

    // the movement methods:

    public void up(){
      viewImagRange[0] += viewImagDif*0.25;
      viewImagRange[1] += viewImagDif*0.25;
      canvas.repaint();
    }

    public void down(){
      viewImagRange[0] -= viewImagDif*0.25;
      viewImagRange[1] -= viewImagDif*0.25;
      canvas.repaint();      
     }

    public void left(){
      viewRealRange[0] -= viewRealDif*0.25;
      viewRealRange[1] -= viewRealDif*0.25;
      canvas.repaint();
     }

    public void right(){
      viewRealRange[0] += viewRealDif*0.25;
      viewRealRange[1] += viewRealDif*0.25;
      canvas.repaint();
     }

    public void zoomIn(){
      viewRealDif /= 2;
      viewImagDif /= 2;
      viewRealRange[0] += viewRealDif*0.5;
      viewRealRange[1] -= viewRealDif*0.5;
      viewImagRange[0] += viewImagDif*0.5;
      viewImagRange[1] -= viewImagDif*0.5; 
      canvas.repaint();
     }

    public void zoomOut(){
      viewRealDif *= 2;
      viewImagDif *= 2;
      viewRealRange[0] -= viewRealDif*0.25;
      viewRealRange[1] += viewRealDif*0.25;
      viewImagRange[0] -= viewImagDif*0.25;
      viewImagRange[1] += viewImagDif*0.25;
      canvas.repaint();
     }

  }


  /** sets the affine transformation, determined by window-size and function
      range */
  public void determineTransform(int width, int height){
    // scale the Image to the appropriate size
    double factorx = ((double) width) / viewRealDif;
    double factory = ((double) height) / viewImagDif;
    
    transform = new AffineTransform(factorx, 0,
			       0, - factory,
			       - viewRealRange[0] * factorx, 
			       viewImagRange[1] * factory);
}

  /** returns a Complex number with points which have been transformed
      from integer-physical graphics coordinates to double logical space
      coordinates  using the inverse Transformation of the one used in
      paint() */
  public Complex transformXY(int x, int y){
    double[] in = {x,y};
    double[] out = new double[2];
    try {
      transform.createInverse().transform(in,0,out,0,1);
    } catch(Exception ex){ex.printStackTrace();}
    return new Complex(out[0], out[1]);
  }


  /** this inner class represents the canvas on which everything is
      drawn. We use physical coordinates (depending on the size of the 
      component) and logical coordinates (depending on the given definition 
      range) which are transformed by methods of the outer class */

  public class GridCanvas extends Canvas {
    public void paint(Graphics g){
      if(n == 0 || m == 0)
	return;
      
      // we use the Java2D-API
      Graphics2D g2 = (Graphics2D) g;
      //g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON); -- too slow!

      // set transform 
      determineTransform(getSize().width,
			 getSize().height);

      g2.setTransform(transform);	

      // set line width to one physical pixel (in logical coordinates!)
      g2.setStroke(new BasicStroke((float)(1 * viewRealDif / 
					   getSize().width), 
				   BasicStroke.CAP_BUTT, 
				   BasicStroke.JOIN_MITER));
      for(int i = 0; i < n; i++)
	for(int j = 0; j < m; j++){	  
		
	  // paint a line, that connects each node to it's 4 neighbours
	  // thus creating a contorted grid
	  if(complexValues[i][j].isValid()){
	    if(i > 0 && complexValues[i-1][j].isValid())
	      g2.draw(new Line2D.Double(complexValues[i][j].real, 
					complexValues[i][j].imag,
					complexValues[i-1][j].real, 
					complexValues[i-1][j].imag));
	    
	    if(i + 1 < n && complexValues[i+1][j].isValid())
	      g2.draw(new Line2D.Double(complexValues[i][j].real, 
					complexValues[i][j].imag,
					complexValues[i+1][j].real, 
					complexValues[i+1][j].imag));
	    
	    if(j > 0 && complexValues[i][j-1].isValid())
	      g2.draw(new Line2D.Double(complexValues[i][j].real, 
					complexValues[i][j].imag,
					complexValues[i][j-1].real, 
					complexValues[i][j-1].imag));
	    
	    if(j + 1 < m && complexValues[i][j+1].isValid())
	      g2.draw(new Line2D.Double(complexValues[i][j].real, 
					complexValues[i][j].imag,
					complexValues[i][j+1].real, 
					complexValues[i][j+1].imag));
	  }
	}
      // draw coordinates
      g2.setColor(Color.red);

      g2.setStroke(new BasicStroke((float) (2 * viewImagDif /getSize().height),
				   BasicStroke.CAP_BUTT, 
				   BasicStroke.JOIN_MITER));
      Line2D.Double myLine = new Line2D.Double(viewRealRange[0], 0, viewRealRange[1], 0);
      g2.draw(myLine);
      g2.setStroke(new BasicStroke((float) (2 * viewRealDif / getSize().width),
				   BasicStroke.CAP_BUTT, 
				   BasicStroke.JOIN_MITER));
      myLine = new Line2D.Double(0, viewImagRange[0], 0, viewImagRange[1]);
      g2.draw(myLine);

    }
  }

  /** returns for a given function value f the approximate origin value 
      (i.e. the value in the grid whose function value best matches f) */
  public Complex getOriginValue(Complex o){
    int min_i = 0, min_j = 0;
    double minimumDistance = Double.POSITIVE_INFINITY;
    for(int i = 0; i < n; i++)
      for(int j = 0; j < m; j++){
	double distance = Complex.subtract(complexValues[i][j], o).abs();
	if(distance < minimumDistance){
	  min_i = i;
	  min_j = j;
	  minimumDistance = distance;
	}
      }

    return new Complex(realRange[0] + min_i * cellWidth, 
		       imagRange[0] + min_j * cellHeight);
  }
	

  /** an inner class to read out mouse coordinates and calculate the
      corresponding values */
  
  public class MouseCoordinateReader extends MouseAdapter implements MouseMotionListener {
    
    public void mouseMoved(MouseEvent e){

      // read out mouse-coordinates (modified by offsets) and apply inverse 
      // transformation on them
      Complex c = transformXY(e.getX(), e.getY());

      // the retrieved coordinates are already the function values!
      double f_x = c.real;     
      double f_y = c.imag;

      // determine the (approximate) origin values
      c = getOriginValue(c);

      // if numbers are of form xxxE-xx (i.e. too low), they must be set to 
      // zero for displaying purposes
      if(Math.abs(c.real) < 0.001d)
	c.real = 0.0d;
      if(Math.abs(c.imag) < 0.001d)
	c.imag = 0.0d;
      if(Math.abs(f_x) < 0.001d)
	f_x = 0.0d;
      if(Math.abs(f_y) < 0.001d)
	f_y = 0.0d;	

      // here we also have to add some Code here for numbers of form xxxExx 
      // (too high)

      // beautify all results by cropping the numbers to 2 digits after '.'
      String outx = Double.toString(c.real);
      String outy = Double.toString(c.imag);
      String outf_x = Double.toString(f_x);
      String outf_y = Double.toString(f_y);
      if(outx.indexOf(".") + 3 < outx.length()) 
	outx = outx.substring(0, outx.indexOf(".") + 3);      
      if(outy.indexOf(".") + 3 < outy.length()) 
	outy = outy.substring(0, outy.indexOf(".") + 3);
      if(outf_x.indexOf(".") + 3 < outf_x.length()) 
	outf_x = outf_x.substring(0, outf_x.indexOf(".") + 3);      
      if(outf_y.indexOf(".") + 3 < outf_y.length()) 
	outf_y = outf_y.substring(0, outf_y.indexOf(".") + 3);


      showCoords1.setText("f(z) = ("+outf_x+","+outf_y+"i)");
      showCoords2.setText("z = ("+outx+","+outy+"i)");      
    }
    
    public void mousePressed(MouseEvent e){}
    public void mouseDragged(MouseEvent e){}
    public void mouseReleased(MouseEvent e){}
  }
	
  public GridDisplay(){
    // initialize, so that paint() doesn't complain
    Complex[][] nullValues = {{}};
    setValues(nullValues);

    //remove MouseListener inserted by subclass and insert own MouseListener
    canvas.removeMouseMotionListener(mouseCoordinateReader);
    canvas.removeMouseListener(mouseCoordinateReader);
    setLayout(new BorderLayout());

    //to read out Complex Coordinates
    canvas.addMouseMotionListener(mouseCoordinateReader);
    canvas.addMouseListener(mouseCoordinateReader);
    setLayout(new BorderLayout());
    add(canvas, BorderLayout.CENTER);
    add(dirPanel, BorderLayout.SOUTH);
    //add(showCoords, BorderLayout.SOUTH);

  }

  /** sets the ranges for both real and imaginary direction */
  public void setRange(double[] realRange, double[] imagRange){
    this.realRange[0] = realRange[0];
    this.realRange[1] = realRange[1];
    this.imagRange[0] = imagRange[0];
    this.imagRange[1] = imagRange[1];
    this.viewRealRange[0] = realRange[0];
    this.viewRealRange[1] = realRange[1];
    this.viewImagRange[0] = imagRange[0];
    this.viewImagRange[1] = imagRange[1];
    realDif = viewRealRange[1] - realRange[0];
    imagDif = viewImagRange[1] - imagRange[0];
    viewRealDif = realDif;
    viewImagDif = imagDif;
  }

  public void setValues(Complex[][] inValues){
    n = inValues.length;
    m = inValues[0].length;
    
    complexValues = new Complex[n][m];
    for(int i = 0; i < n; i++)
      for(int j = 0; j < m; j++)
	complexValues[i][j] = new Complex(inValues[i][j]);
    determineCellSize(); 
  }
}
	
